"""
video_ql utils module.
"""

import base64
import hashlib
import io

import cv2
import numpy as np
from PIL import Image


def video_hash(video_path: str) -> str:
    """Generate a hash for the video file"""
    file_hash = hashlib.md5()
    with open(video_path, "rb") as f:
        # Read in chunks to handle large files
        for chunk in iter(lambda: f.read(4096), b""):
            file_hash.update(chunk)
    return file_hash.hexdigest()


def get_length_of_video(video_path: str) -> int:
    """Get the number of frames in a video by iterating through all frames"""
    cap = cv2.VideoCapture(video_path)
    if not cap.isOpened():
        raise ValueError(f"Could not open video file: {video_path}")

    frame_count = 0
    while True:
        ret, _ = cap.read()
        if not ret:
            break
        frame_count += 1

    cap.release()
    return frame_count


def get_video_fps(video_path: str) -> float:
    """Get the frames per second (FPS) of a video"""
    cap = cv2.VideoCapture(video_path)
    if not cap.isOpened():
        raise ValueError(f"Could not open video file: {video_path}")

    fps = cap.get(cv2.CAP_PROP_FPS)

    cap.release()
    return fps


def encode_image(image_array: np.ndarray) -> str:
    """Encode image array to base64 string."""
    image = Image.fromarray(cv2.cvtColor(image_array, cv2.COLOR_BGR2RGB))
    buffer = io.BytesIO()
    image.save(buffer, format="JPEG")
    buffer.seek(0)
    image_bytes = buffer.getvalue()
    return base64.b64encode(image_bytes).decode("utf-8")



"""Visualization module for video_ql."""

from typing import Any, Dict, List, Tuple

import cv2
import numpy as np

from .models import Label


class VideoVisualizer:
    """Class responsible for video visualization in video_ql."""

    @staticmethod
    def visualize_results(frame: np.ndarray, analysis: Label) -> np.ndarray:
        """
        Overlay analysis results on the frame with a clean, professional look.
        """
        # Create a copy of the frame
        vis_frame = frame.copy()
        h, w = vis_frame.shape[:2]

        # Format results for display
        status_info = {}
        for key, value in analysis.results.items():
            if key != "timestamp":
                # Convert snake_case to Title Case
                formatted_key = key.replace("_", " ").title()
                status_info[formatted_key] = value

        # Create a blue semi-transparent box in the top-left corner
        box_height = 30 * (
            len(status_info) + 1
        )  # Height based on number of items
        box_width = int(0.9 * w)  # Fixed width

        # Create blue box with transparency
        overlay = vis_frame.copy()
        cv2.rectangle(
            overlay,
            (10, 10),
            (10 + box_width, 10 + box_height),
            (255, 0, 0),
            -1,
        )
        vis_frame = cv2.addWeighted(overlay, 0.8, vis_frame, 0.2, 0)

        # Add text to the box in white
        y_pos = 40
        for key, value in status_info.items():
            text = f"{key}: {value}"
            cv2.putText(
                vis_frame,
                text,
                (20, y_pos),
                cv2.FONT_HERSHEY_SIMPLEX,
                0.7,
                (255, 255, 255),
                2,
            )
            y_pos += 30

        # Add timestamp at the bottom
        cv2.putText(
            vis_frame,
            f"Time: {analysis.timestamp:.2f}s",
            (10, vis_frame.shape[0] - 20),
            cv2.FONT_HERSHEY_SIMPLEX,
            0.6,
            (255, 255, 255),
            1,
        )

        # Add error message if present
        if analysis.error:
            cv2.putText(
                vis_frame,
                f"Error: {analysis.error}",
                (10, h - 40),
                cv2.FONT_HERSHEY_SIMPLEX,
                0.6,
                (0, 0, 255),
                1,
            )

        return vis_frame

    @staticmethod
    def create_tile_image_from_frames(
        frames: list, tile_shape: tuple
    ) -> np.ndarray:
        """Create a tiled image from multiple frames."""
        if not frames:
            return np.zeros((100, 100, 3), dtype=np.uint8)

        # Get dimensions from the first frame
        frame_height, frame_width = frames[0]["frame"].shape[:2]
        rows, cols = tile_shape

        # Calculate the dimensions of the tiled image
        tile_height = rows * frame_height
        tile_width = cols * frame_width

        # Create an empty canvas
        tile_image = np.zeros((tile_height, tile_width, 3), dtype=np.uint8)

        # Fill the canvas with frames
        for i, frame_data in enumerate(frames[: rows * cols]):
            row = i // cols
            col = i % cols

            y_start = row * frame_height
            y_end = y_start + frame_height
            x_start = col * frame_width
            x_end = x_start + frame_width

            tile_image[y_start:y_end, x_start:x_end] = frame_data["frame"]

        return tile_image

    @staticmethod
    def extract_frames(
        video_path: str,
        start_idx: int,
        count: int,
        stride: int,
        max_resolution: Tuple[int, int],
    ) -> List[Dict[str, Any]]:
        """Extract frames from video starting at index"""
        cap = cv2.VideoCapture(video_path)

        if not cap.isOpened():
            raise ValueError(f"Could not open video file: {video_path}")

        video_fps = cap.get(cv2.CAP_PROP_FPS)
        frames = []  # type: ignore

        # Set position to start_idx
        cap.set(cv2.CAP_PROP_POS_FRAMES, start_idx)

        for _ in range(count):
            ret, frame = cap.read()

            if not ret:
                break

            # Resize frame if needed
            if frame is not None:
                h, w = frame.shape[:2]  # type: ignore
                max_w, max_h = max_resolution

                scale = 1.0
                if w > max_w or h > max_h:
                    scale_w = max_w / w
                    scale_h = max_h / h
                    scale = min(scale_w, scale_h)

                if scale < 1.0:
                    new_w, new_h = int(w * scale), int(h * scale)
                    frame = cv2.resize(
                        frame, (new_w, new_h), interpolation=cv2.INTER_AREA
                    )

            timestamp = (start_idx + len(frames)) / video_fps
            frames.append({"frame": frame, "timestamp": timestamp})

            # Stride jump
            for _ in range(stride - 1):
                ret, frame = cap.read()

        cap.release()
        return frames



"""
video_ql base module.
"""

import hashlib
import json
import os
from typing import Any, Dict, List, Optional, Union

import cv2
import numpy as np
import yaml
from langchain_anthropic import ChatAnthropic
from langchain_core.messages import HumanMessage
from langchain_core.output_parsers import JsonOutputParser
from langchain_openai import ChatOpenAI
from pydantic import BaseModel, Field, create_model

from .models import Label, Query, QueryConfig, VideoProcessorConfig
from .query import matches_query
from .utils import encode_image, get_length_of_video, get_video_fps, video_hash
from .visualization import VideoVisualizer

NAME = "video_ql"


class VideoQL:
    def __init__(
        self,
        video_path: str,
        queries: List[Query],
        context: str = "Answer the following",
        video_processor_config: Optional[VideoProcessorConfig] = None,
        cache_dir: str = "~/.cache/video_ql/",
        disable_cache: bool = False,
        model_name: str = "gpt-4o-mini",
        # model_name: str = "claude-3-haiku-20240307",
    ):
        """Initialize the VideoQL instance"""
        self.video_path = video_path
        self.queries = queries
        self.context = context
        self.disable_cache = disable_cache
        self.model_name = model_name

        # Expand the cache directory if it starts with ~
        self.cache_dir = os.path.expanduser(cache_dir)

        # Create default config if not provided
        if video_processor_config is None:
            self.config = VideoProcessorConfig(context=context)
        else:
            self.config = video_processor_config
            if not hasattr(self.config, "context"):
                self.config.context = context

        # Generate a unique hash for this video analysis setup
        self.scene_hash = self._generate_scene_hash()
        self.cache_path = os.path.join(
            self.cache_dir, f"{self.scene_hash}.json"
        )

        # Get video info
        self.num_video_frames = get_length_of_video(video_path)
        self.num_frames_per_tile = (
            self.config.tile_frames[0] * self.config.tile_frames[1]
        )
        self.video_fps = get_video_fps(video_path)
        # Calculate the correct frame stride based on fps adjustment
        self.effective_stride = int(
            self.config.frame_stride * (self.video_fps / self.config.fps)
        )

        # Visualizer
        self.visualizer = VideoVisualizer()

        # Create the frame analysis model
        self.frame_model = self._create_frame_model()
        self.parser = JsonOutputParser(
            pydantic_object=self.frame_model  # type: ignore
        )

        # Load or initialize the cache
        if not os.path.exists(os.path.dirname(self.cache_path)):
            os.makedirs(os.path.dirname(self.cache_path))

        self.__cache = self._load_cache()
        self.prompt = self._create_prompt()

    def _generate_scene_hash(self) -> str:
        """Generate a unique hash for this video analysis setup"""
        # Hash the video file
        v_hash = video_hash(self.video_path)

        # Create a string representing the queries and config
        query_str = json.dumps(
            [q.model_dump() for q in self.queries], sort_keys=True
        )
        config_str = json.dumps(self.config.model_dump(), sort_keys=True)
        context_str = self.context

        # Combine all components and hash
        combined = f"{v_hash}_{query_str}_{config_str}_{context_str}"
        return hashlib.md5(combined.encode()).hexdigest()

    def _load_cache(self) -> Dict[int, Label]:
        """Load the cache from disk"""
        cache: Dict[int, Label] = {}
        if self.disable_cache:
            return cache

        if os.path.exists(self.cache_path):
            try:
                with open(self.cache_path, "r") as f:
                    cache_data = json.load(f)
                for key, value in cache_data.items():
                    cache[int(key)] = Label(**value)
            except Exception as e:
                print(f"Warning: Could not load cache: {e}")
        return cache

    def _save_cache(self):
        """Save the cache to disk"""
        if self.disable_cache:
            return

        cache_data = {k: v.model_dump() for k, v in self.__cache.items()}
        with open(self.cache_path, "w") as f:
            json.dump(cache_data, f, indent=2)

    def _create_frame_model(self) -> BaseModel:
        """Create a Pydantic model based on the queries"""
        field_definitions = {}

        for query in self.queries:
            field_name = query.query.lower().replace("?", "").replace(" ", "_")

            if query.options:
                field_definitions[field_name] = (
                    str,
                    Field(
                        description=f"Context: {self.context}; Query: {query.query}; Choose from: {', '.join(query.options)}"  # noqa
                    ),
                )
            else:
                field_definitions[field_name] = (
                    str,
                    Field(description=query.query),
                )

        # Add timestamp field
        field_definitions["timestamp"] = (
            float,
            Field(description="Timestamp of the frame in seconds"),
        )  # type: ignore

        # Create and return the model
        FrameAnalysis = create_model(
            "FrameAnalysis", **field_definitions
        )  # type: ignore

        return FrameAnalysis

    def _create_prompt(self) -> str:
        """Create a prompt based on the queries"""
        prompt = "Analyze this video frame and provide the following information:\n\n"  # noqa

        for query in self.queries:
            if query.options:
                prompt += f"- {query.query} Choose from: {', '.join(query.options)}\n"  # noqa
            else:
                prompt += f"- {query.query}\n"

        return prompt

    def _analyze_frame(self, frame: Dict[str, Any]) -> Label:
        """Analyze a single frame using the vision model"""
        image_base64 = encode_image(frame["frame"])

        if self.model_name in ["gpt-4o-mini"]:
            model = ChatOpenAI(
                temperature=0.3, model=self.model_name, max_tokens=1024
            )  # type: ignore
        elif self.model_name in ["claude-3-haiku-20240307"]:
            model = ChatAnthropic(
                temperature=0.3, model=self.model_name, max_tokens=1024
            )  # type: ignore
        else:
            raise ValueError(f"Unknown model name: {self.model_name}")

        try:
            msg = model.invoke(
                [
                    HumanMessage(
                        content=[
                            {"type": "text", "text": self.prompt},
                            {
                                "type": "text",
                                "text": self.parser.get_format_instructions(),
                            },
                            {
                                "type": "image_url",
                                "image_url": {
                                    "url": f"data:image/jpeg;base64,{image_base64}"  # noqa
                                },
                            },
                        ]
                    )
                ]
            )

            # Parse the output
            parsed_output = self.parser.parse(msg.content)  # type: ignore

            # Make sure timestamp is included
            if (
                "timestamp" not in parsed_output
                or parsed_output["timestamp"] != frame["timestamp"]
            ):
                parsed_output["timestamp"] = frame["timestamp"]

            return Label(timestamp=frame["timestamp"], results=parsed_output)

        except Exception as e:
            return Label(
                timestamp=frame["timestamp"], results={}, error=str(e)
            )

    def __len__(self) -> int:
        """Return the number of frames that can be processed"""
        stride = self.config.frame_stride * (self.video_fps / self.config.fps)
        return max(0, int(self.num_video_frames - stride))

    def __getitem__(self, idx: int) -> Label:
        """Get the analysis for a specific frame index"""
        if idx < 0 or idx >= len(self):
            raise IndexError(
                f"Index {idx} out of bounds for video with {len(self)} frames"
            )

        # Calculate the cache index using the effective stride
        cache_idx = idx // self.effective_stride

        if cache_idx not in self.__cache:
            # Calculate the actual frame index in the video
            frame_idx = cache_idx * self.effective_stride

            # Extract frames for the tile
            total_frames_needed = (
                self.config.tile_frames[0] * self.config.tile_frames[1]
            )
            frames = self.extract_frames(
                frame_idx, total_frames_needed, self.config.frame_stride
            )

            # Create a tile image from these frames
            tile_image = self.create_tile_image_from_frames(frames)

            # Use the timestamp from the first frame
            first_frame_timestamp = frames[0]["timestamp"] if frames else 0

            frame = {
                "frame": tile_image,
                "timestamp": first_frame_timestamp,
            }

            analysis = self._analyze_frame(frame)
            self.__cache[cache_idx] = analysis
            # Save the updated cache
            self._save_cache()

        return self.__cache[cache_idx]

    def get_frame_analysis(self, timestamp: float) -> Label:
        """Get the frame analysis at a specific timestamp"""
        cap = cv2.VideoCapture(self.video_path)
        if not cap.isOpened():
            raise ValueError(f"Could not open video file: {self.video_path}")

        fps = cap.get(cv2.CAP_PROP_FPS)
        frame_idx = int(timestamp * fps)
        cap.release()

        return self[frame_idx]

    def analyze_video(
        self,
        display: bool = False,
        save_frames: bool = False,
        output_dir: str = "results",
    ):
        """Process the entire video and return all analysis results"""
        results = []

        # Create output directory for frames if needed
        if save_frames:
            frames_dir = os.path.join(output_dir, "frames")
            if not os.path.exists(frames_dir):
                os.makedirs(frames_dir)

        # Get total number of frames to process
        total_frames = len(self)

        for i in range(0, total_frames):
            try:
                analysis = self[i]
                results.append(analysis)
                print(
                    f"Processed frame at timestamp: {analysis.timestamp:.2f}s"
                )

                # If we need to display or save, get the original frame
                if display or save_frames:
                    frames = self.extract_frames(i, 1)
                    if frames:
                        vis_frame = self._visualize_results(
                            frames[0]["frame"], analysis
                        )

                        if display:
                            cv2.imshow("Video Analysis", vis_frame)
                            key = cv2.waitKey(1) & 0xFF
                            if key == 27:  # ESC key
                                break

                        if save_frames:
                            frame_path = os.path.join(
                                frames_dir,
                                f"frame_{i:04d}_{analysis.timestamp:.2f}s.jpg",
                            )
                            cv2.imwrite(frame_path, vis_frame)

            except Exception as e:
                print(f"Error processing frame at index {i}: {e}")
                import traceback

                traceback.print_exc()

        # Close any open windows
        if display:
            cv2.destroyAllWindows()

        return results

    def _visualize_results(
        self, frame: np.ndarray, analysis: Label
    ) -> np.ndarray:
        """
        Overlay analysis results on the frame with a clean,
        professional look.
        """
        return self.visualizer.visualize_results(frame, analysis)

    def create_tile_image_from_frames(
        self, frames: List[Dict[str, Any]]
    ) -> np.ndarray:
        """Create a tiled image from multiple frames."""
        return self.visualizer.create_tile_image_from_frames(
            frames, self.config.tile_frames
        )

    def extract_frames(
        self,
        start_idx: int,
        count: int,
        stride: int = 1,
    ) -> List[Dict[str, Any]]:
        frames = self.visualizer.extract_frames(
            self.video_path,
            start_idx,
            count,
            stride,
            self.config.max_resolution,
        )
        return frames

    def create_tile_image(
        self, start_idx: int = 0, stride: int = 1
    ) -> np.ndarray:
        """Create a tiled image of frames starting from start_idx"""
        frames = self.extract_frames(
            start_idx,
            self.config.tile_frames[0] * self.config.tile_frames[1],
            stride,
        )

        return self.create_tile_image_from_frames(frames)

    def query_video(
        self,
        query_config: Union[str, Dict, QueryConfig],
        output_path: str = "results/query_output.mp4",
    ) -> List[int]:
        """
        Query the video based on a query configuration
        Returns indices of frames that match the query

        Args:
            query_config: Path to a YAML file, a dict, or a QueryConfig object
                containing query configuration
            output_path: Path to save the output video
        """
        # Load query config if it's a path
        if isinstance(query_config, str):
            with open(query_config, "r") as f:
                query_data = yaml.safe_load(f)
                query_config = QueryConfig(**query_data)
        # Convert dict to QueryConfig if needed
        elif isinstance(query_config, dict):
            query_config = QueryConfig(**query_config)

        matching_frames = []

        # Process all frames first (if not already in cache)
        self.analyze_video(display=False, save_frames=False)

        # Now search through cached results
        for idx in sorted(self.__cache.keys()):
            analysis = self.__cache[idx]

            # If the analysis matches any of the queries
            if matches_query(analysis, query_config.queries):

                video_idx_lb = int(idx * self.effective_stride)
                video_idx_ub = int((idx + 1) * self.effective_stride)

                for video_idx in range(video_idx_lb, video_idx_ub):
                    matching_frames.append(video_idx)

        return matching_frames



import pytest
from video_ql.models import Label, QueryCondition, AndCondition, OrCondition
from video_ql.query import matches_query, matches_subquery


@pytest.fixture
def sample_analysis():
    return Label(
        timestamp=1.0,
        results={
            "person_present": True,
            "car_visible": False,
            "weather": "sunny",
            "scene_type": "outdoor",
            "empty": "",
        },
    )


def test_matches_subquery_basic():
    analysis = Label(
        timestamp=1.0,
        results={"person_present": True, "car_visible": False},
    )

    # Test basic truthy value check
    query = QueryCondition(query="person present?")
    assert matches_subquery(analysis, query) is True

    query = QueryCondition(query="car visible?")
    assert matches_subquery(analysis, query) is False

    # Test non-existent field
    query = QueryCondition(query="bicycle visible?")
    assert matches_subquery(analysis, query) is False


def test_matches_subquery_with_options(sample_analysis):
    # Test with options where value matches
    query = QueryCondition(query="weather?", options=["sunny", "cloudy"])
    assert matches_subquery(sample_analysis, query) is True

    # Test with options where value doesn't match
    query = QueryCondition(query="weather?", options=["rainy", "cloudy"])
    assert matches_subquery(sample_analysis, query) is False

    # Test with options for truthy value
    query = QueryCondition(query="person present?", options=["True"])
    assert (
        matches_subquery(sample_analysis, query) is False
    )  # Won't match as True != "True"


def test_matches_subquery_empty_values(sample_analysis):
    # Test with empty field
    query = QueryCondition(query="empty?")
    assert matches_subquery(sample_analysis, query) is False


def test_matches_query_simple():
    analysis = Label(
        timestamp=1.0,
        results={"person_present": True, "scene_type": "outdoor"},
    )

    queries = [QueryCondition(query="person present?")]
    assert matches_query(analysis, queries) is True

    queries = [QueryCondition(query="car visible?")]
    assert matches_query(analysis, queries) is False


def test_matches_query_and_condition():
    analysis = Label(
        timestamp=1.0,
        results={"person_present": True, "scene_type": "outdoor"},
    )

    # Test AND where all conditions match
    queries = [
        AndCondition(
            AND=[
                QueryCondition(query="person present?"),
                QueryCondition(query="scene type?", options=["outdoor"]),
            ]
        )
    ]
    assert matches_query(analysis, queries) is True

    # Test AND where not all conditions match
    queries = [
        AndCondition(
            AND=[
                QueryCondition(query="person present?"),
                QueryCondition(query="scene type?", options=["indoor"]),
            ]
        )
    ]
    assert matches_query(analysis, queries) is False


def test_matches_query_or_condition():
    analysis = Label(
        timestamp=1.0,
        results={"person_present": True, "car_visible": False},
    )

    # Test OR where one condition matches
    queries = [
        OrCondition(
            OR=[
                QueryCondition(query="person present?"),
                QueryCondition(query="car visible?"),
            ]
        )
    ]
    assert matches_query(analysis, queries) is True

    # Test OR where no conditions match
    queries = [
        OrCondition(
            OR=[
                QueryCondition(query="bicycle visible?"),
                QueryCondition(query="truck visible?"),
            ]
        )
    ]
    assert matches_query(analysis, queries) is False


def test_matches_query_nested():
    analysis = Label(
        timestamp=1.0,
        results={
            "person_present": True,
            "car_visible": False,
            "weather": "sunny",
            "scene_type": "outdoor",
        },
    )

    # Test OR with nested AND
    queries = [
        OrCondition(
            OR=[
                AndCondition(
                    AND=[
                        QueryCondition(query="person present?"),
                        QueryCondition(query="weather?", options=["rainy"]),
                    ]
                ),
                AndCondition(
                    AND=[
                        QueryCondition(
                            query="scene type?", options=["outdoor"]
                        ),
                        QueryCondition(query="weather?", options=["sunny"]),
                    ]
                ),
            ]
        )
    ]
    assert matches_query(analysis, queries) is True

    # Test where nothing matches
    queries = [
        OrCondition(
            OR=[
                AndCondition(
                    AND=[
                        QueryCondition(query="person present?"),
                        QueryCondition(query="weather?", options=["rainy"]),
                    ]
                ),
                AndCondition(
                    AND=[
                        QueryCondition(
                            query="scene type?", options=["indoor"]
                        ),
                        QueryCondition(query="weather?", options=["sunny"]),
                    ]
                ),
            ]
        )
    ]
    assert matches_query(analysis, queries) is False


def test_matches_query_multiple_top_level():
    analysis = Label(
        timestamp=1.0,
        results={"person_present": True, "car_visible": False},
    )

    # Test with multiple top-level queries (only one needs to match)
    queries = [
        QueryCondition(query="bicycle visible?"),
        QueryCondition(query="person present?"),
    ]
    assert matches_query(analysis, queries) is True

    # Test with multiple top-level queries where none match
    queries = [
        QueryCondition(query="bicycle visible?"),
        QueryCondition(query="truck visible?"),
    ]
    assert matches_query(analysis, queries) is False



"""CLI interface for video_ql project."""

import argparse
import json
import os

import cv2
import yaml

from .base import VideoQL
from .models import Query, QueryConfig, VideoProcessorConfig


def parse_args():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        description="Query video analysis results"
    )
    parser.add_argument(
        "--video", type=str, required=True, help="Path to the video file"
    )
    parser.add_argument(
        "--config",
        type=str,
        required=True,
        help="Path to the YAML config file",
    )
    parser.add_argument(
        "--results",
        type=str,
        required=True,
        help="Path to the results JSON file",
    )
    parser.add_argument(
        "--query", type=str, required=True, help="Path to the query YAML file"
    )
    parser.add_argument(
        "--output",
        type=str,
        default="results/query_results",
        help="Path to output directory",
    )
    parser.add_argument(
        "--display",
        action="store_true",
        help="Display frames that match the query",
    )
    return parser.parse_args()


def load_config(config_path: str) -> dict:
    """Load configuration from YAML file."""
    with open(config_path, "r") as f:
        config = yaml.safe_load(f)
    return config


def main():  # pragma: no cover
    args = parse_args()
    config = load_config(args.config)

    video_processor_config = VideoProcessorConfig(
        **{
            "fps": config.get("fps"),
            "tile_frames": config.get("tile_frames"),
            "frame_stride": config.get("frame_stride"),
            "max_resolution": config.get("max_resolution"),
        }
    )

    # Convert config queries to Query objects
    queries = [
        Query(
            query=q["query"],
            options=q.get("options"),
            short_query=q.get("short_query"),
            short_options=q.get("short_options"),
        )
        for q in config["queries"]
    ]

    # Initialize VideoQL
    video_ql = VideoQL(
        video_path=args.video,
        queries=queries,
        context=config.get("context", "Answer the following"),
        # The rest of the config options
        video_processor_config=video_processor_config,
    )

    # Load the query
    query_data = load_config(args.query)
    query_config = QueryConfig(**query_data)

    # Find matching frames
    matching_frames = video_ql.query_video(query_config)

    # Create output directory if it doesn't exist
    if not os.path.exists(args.output):
        os.makedirs(args.output)

    # Save and/or display matching frames
    if matching_frames:
        print(f"Found {len(matching_frames)} matching frames")

        # Save information about matching frames
        with open(os.path.join(args.output, "query_results.json"), "w") as f:
            json.dump(
                {
                    "query_config": query_config.model_dump(),
                    "matching_frames": [
                        {
                            "index": idx,
                            "timestamp": video_ql[
                                idx // config.get("frame_stride", 9)
                            ].timestamp,
                        }
                        for idx in matching_frames
                    ],
                },
                f,
                indent=2,
            )

        # Display or save matching frames
        for i, idx in enumerate(matching_frames):
            analysis = video_ql[idx]

            # Extract the frame
            frames = video_ql.extract_frames(idx, 1)
            if frames:
                frame = frames[0]["frame"]
                vis_frame = video_ql._visualize_results(frame, analysis)

                # Save the frame
                frame_path = os.path.join(
                    args.output,
                    f"match_{i:03d}_frame_{idx:04d}_{analysis.timestamp:.2f}s.jpg",  # noqa
                )
                cv2.imwrite(frame_path, vis_frame)

                # Display if requested
                if args.display:
                    cv2.imshow("Query Results", vis_frame)
                    key = cv2.waitKey(1) & 0xFF
                    if key == 27:  # ESC key
                        break

        if args.display:
            cv2.destroyAllWindows()
    else:
        print("No frames matched the query")
